/*
 * (c) 1998,99 Gerd Knorr
 *    capture a image, compress as jpeg and upload to the webserver
 *    using ftp
 *
 * SAO RAS changes:
 *     2001 Vch replace ftp-process with operations from library ftplib.a
 *     2001 VSher have included:
 *                - "grey-scale" mode (-g);          default: color mode
 *                - starting brightness level (-b %) default: 50%
 *                - starting contrast  level (-c %)  default: 30%
 *                - accumulation of images (-n Nadd) default: 10
 *                - statistics on the result image
 *                - determination of bottom/top intensity limits
 *                - scaling image back to 1-byte depth
 *                - brightness/contrast and Nadd auto-adjustment (-a)
 *                - form mini-image (one can use as icon)
 *     2002 VSher have included:
 *                - two TV-channel - fork() for add.chan.service (-s Nchan)
 *                - try to extract&remove line/column structures (-r [Nchan])
 *                - file names now includes "_hostname_Nchan"
 *                - jpeg-files at local Web-site: /home/httpd/html/webcam/
 *                - signature text now includes "/hostname/Nchan"
 *                - force "local mode" - don't upload jpegs to Website (-l)
 *                - force "ftp mode" - upload  to default Website (-f)
 *                - WebSite host address for FTP uploadings (-f host.domain)
 *                - video device name (-d name) default: /dev/video0
 *                - "No TV-signal" detector
 *                - "LiveCamera" interface (2Hz single frame jpegs)
 *     2003 VSher have included:
 *                - WebCamera update time delay (-tw sec)  default: 10.
 *                - LiveCamera update time delay (-tl sec) default: 0.3 (~2.5Hz)
 * Usage:
 *    webcam [-g] [-a] [-b dd.d] [-c dd.d] [-n dd] [-s d] [-r [N]] \
 *           [-l] [-f [addr]] [-d vname] [-tw dd.] [-tl 0.d]
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <sys/types.h>
#include <time.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <asm/types.h>
#include <linux/videodev.h>   /* change this to "videodev2.h" for v4l2 */
#include <jpeglib.h>

#include "ftplib.h"

/* ---------------------------------------------------------------------- */
/* configuration                                                          */

#define FTP_HOST          "mail.sao.ru"
#define FTP_USER          "webcam"
#define FTP_PASS          "webcam@sao.ru"
#define FTP_DIRECTORY     "WWW"
#define FTP_FILE          "webcam_%s_%d.jpeg"
static char ftp_file[100];
#define FTP_FILE_MINI     "webcam_%s_%d_mini.jpeg"
static char ftp_file_mini[100];
#define FTP_TMPFILE       "upload_%s.jpeg"
static char ftp_tmpfile[100];
#define FTP_INIT_ONCE     "pass"		/* call once at start */
#define FTP_INIT_CONN     "bin"			/* per connection */

#define LOCK_FILE         "/tmp/webcam.lock"
//#define JPEG_DIRECTORY    "/home/httpd/html/webcam/"      /*for RH6.2 */
#define JPEG_DIRECTORY    "/var/www/html/webcam/"           /*for RH7.3|RH8.0*/
//#define JPEG_DIRECTORY    "./"                            /*for tests */
static char jpeg_file[100];
static char jpeg_file_mini[100];
#define LIVE_FILE          "livecam_%d.jpeg"
static char live_file[100];
#define JPEG_QUALITY      75
#define LIVE_JPEG_QUALITY 50

#define GRAB_DEVICE       "/dev/video0"
#define GRAB_WIDTH        324
#define GRAB_HEIGHT       243
//#define GRAB_WIDTH        400
//#define GRAB_HEIGHT       300
//#define GRAB_WIDTH        640
//#define GRAB_HEIGHT       480
#define GRAB_NADD         10
#define NADD_MAX_GREY     255
#define NADD_MAX_COLR     85
#define NADD_MIN_GREY     5
#define NADD_MIN_COLR     2
#define GRAB_TEXT_HDR     "webcam/%s/%d:"          /*... hostname/chan: */
#define GRAB_TEXT         " SAO RAS, BTA %d.%m.%Y %H:%M:%S" /* strftime */
static char grab_text[100];
#define GRAB_TEXT_MINI    "%s/%d: BTA SAO RAS"   /* hostname/chan: .... */

#ifdef VIDIOCGCAP
/* these work for v4l only, not v4l2 */
# define GRAB_SOURCE      1
# define GRAB_NORM        VIDEO_MODE_PAL
#endif

static double dt_web=10.;             /* 10sec webcam */
static double dt_live=0.3;            /* ~2.5Hz livecam */
#define GRAB_DELAY        dt_web

/* ---------------------------------------------------------------------- */

void swap_rgb24(char *mem, int n);
static void copy_image(unsigned char *image, int width, int height);
static void add_image(unsigned char *image, int width, int height);
static void out_image(unsigned char *image, int width, int height);
static void make_mini_image(unsigned char *image, int width, int height, int fact, unsigned char *mini_image);
static void prep_frame(unsigned char *image, int width, int height);
static void stat_image(int width, int height);
static int  check_tv_signal();
static void adjust_video_control();
static void lin_col_bgd0();
static void lin1_col1_bgd(int width, int height);
static void dline1_bgd(double *line, int len, int wnd, double *bgd);
static void iline1_bgd(unsigned short *line, int len, int wnd, int *bgd);
static void lin_col_bgd(int width, int height, int hwnd, int vwnd);
static void dsleep(double dt);
static double dtime();
static int color_mode=1;
static int nadd=GRAB_NADD;

/* ---------------------------------------------------------------------- */
/* jpeg stuff                                                             */

int
write_jpeg(char *filename, char *data, int width, int height, int jpeg_quality)
{
    struct jpeg_compress_struct cinfo;
    struct jpeg_error_mgr jerr;
    FILE *fp;
    int i,nc,fd;
    unsigned char *line;

    if(jpeg_quality<10 || jpeg_quality>100) jpeg_quality = JPEG_QUALITY;

    if (NULL == (fp = fopen(filename,"w"))) {
	fprintf(stderr,"can't open %s for writing: %s\n",
		filename,strerror(errno));
	return -1;
    } else fd = fileno(fp);
    if(flock(fd, LOCK_EX)<0) perror("locking jpeg-file");

    nc = color_mode? 3 : 1;

    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_compress(&cinfo);
    jpeg_stdio_dest(&cinfo, fp);
    cinfo.image_width = width;
    cinfo.image_height = height;
    cinfo.input_components = nc;
    cinfo.in_color_space = color_mode? JCS_RGB : JCS_GRAYSCALE;
    jpeg_set_defaults(&cinfo);
    jpeg_set_quality(&cinfo, jpeg_quality, TRUE);
    jpeg_start_compress(&cinfo, TRUE);

    for (i = 0, line = data; i < height; i++, line += width*nc)
	jpeg_write_scanlines(&cinfo, &line, 1);
    
    jpeg_finish_compress(&(cinfo));
    jpeg_destroy_compress(&(cinfo));

    if(flock(fd, LOCK_UN)<0) perror("unlocking jpeg-file");
    fclose(fp);

    return 0;
}

/* ---------------------------------------------------------------------- */
/* FTP stuff                                                              */
static netbuf *ftp = NULL;
static int ftp_connected=0;
void ftp_connect(char *ftp_host)
{
 if (!FtpConnect(ftp_host,&ftp))  {
     fprintf(stderr,"Unable to connect to node %s\n",ftp_host);
     exit(1);
 }
 if (!FtpLogin(FTP_USER,FTP_PASS,ftp)) {
     fprintf(stderr,"Login failure\n");
     exit(1);
 }
 ftp_connected=1;
 if (!FtpChdir(FTP_DIRECTORY, ftp))  {
     fprintf(stderr,"Change directory failed\n");
     exit(1);
 }
 FtpOptions(FTPLIB_IDLETIME, (long) 1000, ftp);
}

void ftp_upload(char *local_file, char *remote_file)
{
 FtpPut(local_file, ftp_tmpfile, FTPLIB_BINARY, ftp);
 FtpRename(ftp_tmpfile, remote_file, ftp);
//FtpClose(ftp);
}


/* ---------------------------------------------------------------------- */
/* capture stuff  - v4l2                                                  */

#ifdef VIDIOC_QUERYCAP

static struct v4l2_capability    grab_cap;
static struct v4l2_format        grab_pix;
static int                       grab_fd, grab_size;
static unsigned char            *grab_data;

void
grab_init()
{
    if (-1 == (grab_fd = open(GRAB_DEVICE,O_RDWR))) {
	perror("open " GRAB_DEVICE);
	exit(1);
    }
    if (-1 == ioctl(grab_fd,VIDIOC_QUERYCAP,&grab_cap)) {
	fprintf(stderr,"wrong device\n");
	exit(1);
    }
    if (-1 == ioctl(grab_fd, VIDIOC_G_FMT, &grab_pix)) {
        perror("ioctl VIDIOC_G_FMT");
	exit(1);
    }
    grab_pix.pixelformat = V4L2_PIX_FMT_BGR24;
    grab_pix.depth  = 24;
    grab_pix.width  = GRAB_WIDTH;
    grab_pix.height = GRAB_HEIGHT;
    if (-1 == ioctl(grab_fd, VIDIOC_S_FMT, &grab_pix)) {
	perror("ioctl VIDIOC_S_FMT");
	exit(1);
    }
    grab_size = grab_pix.width * grab_pix.height * ((grab_pix.depth+7)/8);
    fprintf(stderr,"grabber: using %dx%dx%d => %d byte\n",
	    grab_pix.width,grab_pix.height,grab_pix.depth,grab_size);
    if (NULL == (grab_data = malloc(grab_size))) {
	fprintf(stderr,"out of virtual memory\n");
	exit(1);
    }
}

unsigned char*
grab_one(int *width, int *height)
{
    int rc;

    for (;;) {
	rc = read(grab_fd,grab_data,grab_size);
	if (rc == grab_size) {
	    swap_rgb24(grab_data,grab_pix.width*grab_pix.height);
	    *width  = grab_pix.width;
	    *height = grab_pix.height;
	    return grab_data;
	}
	fprintf(stderr,"grabber: read: %d != %d\n",grab_size,rc);
	if (-1 == rc)
	    perror("grabber: read");
    }
}

#endif


/* ---------------------------------------------------------------------- */
/* capture stuff  -  old v4l (bttv)                                       */

#ifdef VIDIOCGCAP

static struct video_capability   grab_cap;
static struct video_mbuf         grab_buffers;
static struct video_mmap         grab_buff[2];
static struct video_channel	 grab_chan;
static struct video_picture      grab_pict;
static double brightness=50.0, contrast=30.0; /* 0.0% - 100.0% */
static int  grab_fd=-1, grab_size, buf_cnt, grab_need_sync=0;
static unsigned int grab_cnt;
static unsigned char *grab_data;

static void grab_init(char *dev)
{
    int i;
    if (-1 == (grab_fd = open(dev,O_RDWR))) {
	char tmp[100];
	snprintf(tmp,100,"open %s",dev);
	perror(tmp);
	exit(1);
    }
    if (-1 == ioctl(grab_fd,VIDIOCGCAP,&grab_cap)) {
	fprintf(stderr,"wrong device\n");
	exit(1);
    }

    /* set image source and TV norm */
    grab_chan.channel = GRAB_SOURCE;
    if (-1 == ioctl(grab_fd,VIDIOCGCHAN,&grab_chan)) {
	perror("ioctl VIDIOCGCHAN");
	exit(1);
    }
    grab_chan.channel = GRAB_SOURCE;    
    grab_chan.norm    = GRAB_NORM;
    if (-1 == ioctl(grab_fd,VIDIOCSCHAN,&grab_chan)) {
	perror("ioctl VIDIOCSCHAN");
	exit(1);
    }
    /* set new image properties */
    if (ioctl(grab_fd,VIDIOCGPICT,&grab_pict) < 0 ) {
	perror("ioctl VIDIOCGPICT");
	exit(1);
    }
    grab_pict.brightness = (int)(brightness/100.*65535);
    grab_pict.contrast = (int)(contrast/100.*65535);
    if(color_mode)
       grab_pict.palette = VIDEO_PALETTE_RGB24;
    else
       grab_pict.palette = VIDEO_PALETTE_GREY;
    if (ioctl(grab_fd,VIDIOCSPICT,&grab_pict) < 0 ) {
	perror("ioctl VIDIOCSPICT");
	exit(1);
    }

    /* map grab buffer */
    if (ioctl(grab_fd,VIDIOCGMBUF,&grab_buffers) < 0) {
	perror("ioctl VIDIOCGMBUF");
	exit(1);
    }
    grab_data = mmap(0,grab_buffers.size,PROT_READ|PROT_WRITE,MAP_SHARED,grab_fd,0);
    if ((int)grab_data == -1) {
	perror("mmap");
	exit(1);
    }
    buf_cnt = (grab_buffers.frames <= 2)? grab_buffers.frames : 2;
    grab_cnt = 0;
    for(i=0; i<buf_cnt; i++) {
	if(color_mode)
	   grab_buff[i].format = VIDEO_PALETTE_RGB24;
	else
	   grab_buff[i].format = VIDEO_PALETTE_GREY;
	grab_buff[i].frame  = i;
	grab_buff[i].width  = (GRAB_WIDTH<grab_cap.maxwidth)? GRAB_WIDTH : grab_cap.maxwidth;
	grab_buff[i].height = (GRAB_HEIGHT<grab_cap.maxheight)? GRAB_HEIGHT : grab_cap.maxheight;
    }
    grab_size = grab_buff[0].width * grab_buff[0].height;
    if(color_mode) grab_size *= 3;
}

static int  lock_fd=-1;
static void grab_lock() {
    if(lock_fd<0) {
	if ((lock_fd = open(LOCK_FILE, O_RDWR|O_CREAT, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH))<0) {
	    perror("open " LOCK_FILE);
	    exit(1);
	}
    }
    if(flock(lock_fd, LOCK_EX)<0) perror("locking " LOCK_FILE);
}

static void grab_set_chan(int ch_num)
{
    /* get image source and TV norm */
    grab_chan.channel = ch_num;
    if (ioctl(grab_fd,VIDIOCGCHAN,&grab_chan) < 0 ) {
	perror("ioctl VIDIOCGCHAN");
	exit(1);
    }
    /* set image source and TV norm */
    grab_chan.channel = ch_num;
    grab_chan.norm    = GRAB_NORM;
    if (ioctl(grab_fd,VIDIOCSCHAN,&grab_chan) < 0 ) {
	perror("ioctl VIDIOCSCHAN");
	exit(1);
    }

    /* set chan image properties */
    grab_pict.brightness = (int)(brightness/100.*65535);
    grab_pict.contrast = (int)(contrast/100.*65535);
    if(color_mode)
       grab_pict.palette = VIDEO_PALETTE_RGB24;
    else
       grab_pict.palette = VIDEO_PALETTE_GREY;
    if (ioctl(grab_fd,VIDIOCSPICT,&grab_pict) < 0 ) {
	perror("ioctl VIDIOCSPICT");
	exit(1);
    }
}

static void grab_free() {
    if(flock(lock_fd, LOCK_UN)<0) perror("unlocking" LOCK_FILE);
}

static void grab_adjust_brightness(double adjval)
{
    if(adjval>0.1 && brightness<69.9) {
       brightness += adjval;
       if(brightness>70.) brightness=70.;
    } else if(adjval<-0.1 && brightness>10.1) {
       brightness += adjval;
       if(brightness<10.) brightness=10.;
    } else return;

    grab_pict.brightness = (int)(brightness/100.*65535);
    fprintf(stderr,"Adjust brightness to %5.1f%%  ",brightness);

    if (ioctl(grab_fd,VIDIOCSPICT,&grab_pict) < 0 ) {
	perror("ioctl VIDIOCSPICT");
	exit(1);
    }
}

static void grab_adjust_contrast(double adjval)
{
    if(adjval>0.1 && contrast<69.9) {
       contrast += adjval;
       if(contrast>70.) contrast=70.;
    } else if(adjval<-0.1 && contrast>10.1) {
       contrast += adjval;
       if(contrast<10.) contrast=10.;
    } else return;

    grab_pict.contrast = (int)(contrast/100.*65535);
    fprintf(stderr,"Adjust contrast to %5.1f%%  ",contrast);

    if (ioctl(grab_fd,VIDIOCSPICT,&grab_pict) < 0 ) {
	perror("ioctl VIDIOCSPICT");
	exit(1);
    }
}

static unsigned char* grab_one(int *width, int *height)
{
    for (;;) {
	if (-1 == ioctl(grab_fd,VIDIOCMCAPTURE,&grab_buff[0])) {
	    perror("ioctl VIDIOCMCAPTURE");
        } else {
	    if (-1 == ioctl(grab_fd,VIDIOCSYNC,&grab_buff[0])) {
		perror("ioctl VIDIOCSYNC");
	    } else {
		swap_rgb24(grab_data+grab_buffers.offsets[0],grab_buff[0].width*grab_buff[0].height);
		*width  = grab_buff[0].width;
		*height = grab_buff[0].height;
		return(grab_data + grab_buffers.offsets[0]);
	    }
        }
	sleep(1);
    }
}

static unsigned char* grab_next(int first, int last, int *width, int *height)
{
    int old_frame = grab_cnt % buf_cnt;
    int new_frame = (grab_cnt+1) % buf_cnt;

    if(grab_fd<0) return(NULL); /*against idiots protection :-)*/

    if( first && grab_need_sync==0) {
	if (ioctl(grab_fd,VIDIOCMCAPTURE,&grab_buff[old_frame]) < 0) {
	    perror("ioctl VIDIOCMCAPTURE");
	    return(NULL);
	}
	grab_need_sync++;
      /*printf("VIDIOCMCAPTURE for grab_buf%d (grab_offset=%d)\n",grab_buff[old_frame].frame, grab_buffers.offsets[old_frame]);*/
    }
    if( !last && grab_need_sync<buf_cnt) {
	if (ioctl(grab_fd,VIDIOCMCAPTURE,&grab_buff[new_frame]) < 0) {
	    perror("ioctl VIDIOCMCAPTURE");
	    return(NULL);
	} else
	    grab_cnt++;
	grab_need_sync++;
      /*printf("VIDIOCMCAPTURE for grab_buf%d (grab_offset=%d)\n",grab_buff[new_frame].frame, grab_buffers.offsets[new_frame]);*/
    }
    if (grab_need_sync>0) {
vidiosync:
	if (ioctl(grab_fd,VIDIOCSYNC,&grab_buff[old_frame]) < 0) {
	   if(errno == EINTR) { /* interrupted?    */
	      perror("ioctl VIDIOCSYNC");
	      sleep(1);
	      goto vidiosync;   /* continue with VIDIOCSYNC */
	   }
	   perror("ioctl VIDIOCSYNC");
	   grab_need_sync=0;
	   return(NULL);
       } else {
	   grab_need_sync--;
	   if(grab_buff[old_frame].format == VIDEO_PALETTE_RGB24)
	       swap_rgb24(grab_data+grab_buffers.offsets[old_frame],grab_buff[old_frame].width*grab_buff[old_frame].height);
	   *width  = grab_buff[old_frame].width;
	   *height = grab_buff[old_frame].height;
	 /*printf("grab_cnt=%d grab_offset=%d\n",grab_cnt, grab_buffers.offsets[old_frame]);*/
	   return(grab_data + grab_buffers.offsets[old_frame]);
       }
    } else {
       fprintf(stderr,"Internal error! No frame for VIDIOCSYNC! Use old one...\n");
       return(grab_data + grab_buffers.offsets[old_frame]);
    }
}

#endif

/* ---------------------------------------------------------------------- */

#if 0
# define CHAR_HEIGHT  8
# define CHAR_WIDTH   8
# define CHAR_START   1
# include "font_8x8.h"
#else
# define CHAR_HEIGHT  11
# define CHAR_WIDTH   6
# define CHAR_START   4
# include "font_6x11.h"
#endif

void
add_text(char *image, int width, int height, char *alt_text)
{
    time_t      t;
    struct tm  *tm;
    char        line[128];
    unsigned int cb[128];
    unsigned char *ptr,c;
    int         i,x,y,f,len,k,nc;

    if(alt_text==NULL) {
       time(&t);
       tm = localtime(&t);
       len = strftime(line,127,grab_text,tm);
    } else {
       strncpy(line,alt_text,127);
       len = strlen(line);
    }
    /*fprintf(stderr,"%s\n",line);*/

    if((len+1)*CHAR_WIDTH > width) len = width/CHAR_WIDTH-1;

    nc = color_mode? 3 : 1;

    /* background under line characters */
    for (x = 0; x < len; x++)
	cb[x]=0;
    for (y = height-CHAR_HEIGHT-3; y < height-1; y++)
	for (x = 0; x < len; x++)
	    for (i = 0; i<CHAR_WIDTH*nc; i++) {
		c = image[(y*width + (x+1)*CHAR_WIDTH)*nc + i];
		cb[x] += c;
	    }
    for (x = 1; x < len-1; x++)
	cb[x] = (cb[x-1]+cb[x]+cb[x+1])/3;
    for (x = 0; x < len; x++)
	cb[x] /= (CHAR_HEIGHT+2)*CHAR_WIDTH*nc;

    /* draw line characters */
    for (y = 0; y < CHAR_HEIGHT; y++) {
	ptr = image + nc * width * (height-CHAR_HEIGHT-2+y) + CHAR_WIDTH;
	for (x = 0; x < len; x++) {
	    f = fontdata[line[x] * CHAR_HEIGHT + y];
	    for (i = CHAR_WIDTH-1; i >= 0; i--) {
		if (f & (CHAR_START << i)) {
		    for(k=0;k<nc;k++)
		       ptr[k] = (cb[x]>127)? 0 : 255;
		} else {
		    for(k=0;k<nc;k++)
		       ptr[k] = (int)(ptr[k]-127)/2 + 127;
		}
		ptr += nc;
	    }
	}
    }
}

void
swap_rgb24(char *mem, int n)
{
    char  c;
    char *p = mem;
    int   i = n;
    
    while (--i) {
	c = p[0]; p[0] = p[2]; p[2] = c;
	p += 3;
    }
}

static unsigned char  buff[GRAB_WIDTH*GRAB_HEIGHT*3];
static unsigned char  buff_mini[GRAB_WIDTH*GRAB_HEIGHT*3/4];
static int do_bgd=0;
static int no_tv_signal=0;

int
main(int argc, char *argv[])
{
    unsigned char *image;
    int n, width,height, prev_nadd, pid;
    char hname[80], str[80], tmp[80];
    char *hostname=NULL;
    FILE *fd;
    char *grab_dev=GRAB_DEVICE;
    int auto_mode=0;
    int nadd_mode=0;
    int ftp_send=0;
    char *ftp_host=FTP_HOST;
    int nchan=GRAB_SOURCE;
    int naltch=GRAB_SOURCE;
    int rbgd[3] = {0,0,0};
    double t_liv, t0_add, t0, t, t_start;

    color_mode = 1;
    if(strstr(argv[0],"grey"))
	color_mode = 0;
    for(n=1; n<argc; n++) {
       if (argv[n][0]=='-') {
	  if (argv[n][1]=='g')
	     color_mode = 0;
	  else if (argv[n][1]=='a')
	     auto_mode = 1;
	  else if (argv[n][1]=='b') {
	     n++;
	     if(n<argc) brightness = atof(argv[n]);
	  } else if (argv[n][1]=='c') {
	     n++;
	     if(n<argc) contrast = atof(argv[n]);
	  } else if (argv[n][1]=='n') {
	     n++;
	     if(n<argc) nadd = atoi(argv[n]);
	     nadd_mode=1;
	  } else if (argv[n][1]=='s') {
	     n++;
	     if(n<argc) nchan = atoi(argv[n]);
	     if(nchan<0||nchan>2) nchan=GRAB_SOURCE;
	  } else if (argv[n][1]=='r') {
	     int nc=GRAB_SOURCE;
	     n++;
	     if(n<argc) {
		if (argv[n][0]=='-') n--;
		else nc = atoi(argv[n]);
	     }
	     if(nc>=0 && nc<3) rbgd[nc]=1;
	  } else if (argv[n][1]=='l') {
	     ftp_send=0;
	  } else if (argv[n][1]=='f') {
	     ftp_send=1;
	     n++;
	     if(n<argc) {
		if (argv[n][0]=='-') n--;
		else   ftp_host = strdup(argv[n]);
	     }
	  } else if (argv[n][1]=='d') {
	     n++;
	     if(n<argc) grab_dev = strdup(argv[n]);
	  } else if (argv[n][1]=='t') {
	     n++;
	     if(n<argc) {
		if (argv[n-1][2]=='l')
		   dt_live = atof(argv[n]);
		else
		   dt_web = atof(argv[n]);
	     }
	  }
       }
    }
    if(auto_mode) nadd_mode=0;
    if(color_mode) {
       if(nadd>NADD_MAX_COLR) nadd=NADD_MAX_COLR;
       if(nadd<NADD_MIN_COLR) nadd=NADD_MIN_COLR;
    } else {
       if(nadd>NADD_MAX_GREY) nadd=NADD_MAX_GREY;
       if(nadd<NADD_MIN_GREY) nadd=NADD_MIN_GREY;
    }
    if((fd=popen("hostname","r"))!=NULL) {
       hostname=fgets(hname,39,fd);
       pclose(fd);
    }
    if(hostname==NULL) {
       char *p;
       if((p=getenv("HOSTNAME"))!=NULL)
	  strcpy(hname,p);
       else
	  strcpy(hname,"X");
    }
    hostname = strtok(hname,".\n");
    sprintf( ftp_file,     FTP_FILE,     hostname,GRAB_SOURCE);
    sprintf( ftp_file_mini,FTP_FILE_MINI,hostname,GRAB_SOURCE);
    sprintf( ftp_tmpfile,  FTP_TMPFILE,  hostname);


    fprintf(stderr,"video4linux webcam v1.1 - (c) 1998,99 Gerd Knorr\n");
    fprintf(stderr,"device:%s  mode:%s\n",grab_dev, color_mode?"color":"grey-scale");
    fprintf(stderr,"grabber config: size %dx%d, jpeg quality %d, delays %.1f,%.2fsec\n",
		   GRAB_WIDTH,GRAB_HEIGHT,JPEG_QUALITY, dt_web,dt_live);
    if(ftp_send)
       fprintf(stderr,"ftp config:\n  " FTP_USER "@%s:" FTP_DIRECTORY
		   "\n   %s => %s\n",ftp_host,ftp_tmpfile,ftp_file);
    
    grab_init(grab_dev);

    if(ftp_send) {
       FtpInit();
       ftp_connect(ftp_host);
    }

    lin_col_bgd0();
    prev_nadd = nadd;

    if(nchan!=GRAB_SOURCE) {  /* two TV-channels? */
       if((pid=fork())==0)
	  fprintf(stderr,"Start additional webcam for chan%d\n",nchan);
       else {
	  if(pid<0)
	     perror("Can't create additional webcam process");
	  naltch=nchan;
	  nchan=GRAB_SOURCE;
       }
    }
    sprintf( grab_text,  GRAB_TEXT_HDR,  hostname,nchan);
    strcat( grab_text,   GRAB_TEXT);
    sprintf( str,        GRAB_TEXT_MINI, hostname,nchan);
    sprintf( ftp_file,     FTP_FILE,     hostname,nchan);
    sprintf( ftp_file_mini,FTP_FILE_MINI,hostname,nchan);
    strcpy(jpeg_file, JPEG_DIRECTORY);
    strcat(jpeg_file, ftp_file);
    strcpy(jpeg_file_mini, JPEG_DIRECTORY);
    strcat(jpeg_file_mini, ftp_file_mini);
    sprintf(live_file, JPEG_DIRECTORY LIVE_FILE, nchan);

    t_start = t = dtime();
    t_liv = t-0.51;
    t0 = t-nadd*0.04;
    t0_add = t-GRAB_DELAY-0.1;

    for (;;) {
	int nn, first=1, last=1;
	grab_lock();
	t = dtime();
	if (ftp_send && !ftp_connected)
	    ftp_connect(ftp_host);
	grab_set_chan(nchan);

	nn = (t-t0_add>=GRAB_DELAY)? nadd : 1;
	for(n=0; n<nn; n++) {
	   int stop_wait=0;
	   first = (n==0 || last);
	   last = (n==nn-1);
	   if(nchan!=naltch && !last &&
	      ((nchan==GRAB_SOURCE && (n%12)==11) ||
	       (nchan!=GRAB_SOURCE && ((n%5)==4 || nadd<11))))
	      stop_wait = last = 1;

#ifdef VIDIOC_QUERYCAP     /* capture stuff  - v4l2 */
	   image = grab_one(&width, &height);
#endif
#ifdef VIDIOCGCAP          /* capture stuff  -  old v4l (bttv) */
	   image = grab_next(first, last, &width, &height);
#endif

	   t = dtime();
	   if(n==0 && (nn==nadd || t-t0_add>=GRAB_DELAY))
	      t0=t;

	   if(n==0) copy_image(image, width, height);
	   else     add_image(image, width, height);

	   if(t-t_liv > dt_live) {
	      prep_frame(image, width, height);
	      add_text(image, width, CHAR_HEIGHT+3, NULL);
	      if(no_tv_signal)
		 add_text(image, width, height/2, "                  No TV signal!");
	      write_jpeg(live_file, image, width, height, LIVE_JPEG_QUALITY);
	      t_liv = t;
//printf("T:%.2f Ch%d Frame%d of %d (form livecam_%d.jpeg)\n",t-t_start,nchan,n,nn,nchan);
	   }
//else printf("T:%.2f Ch%d Frame%d of %d\n",t-t_start,nchan,n,nn);

	   if(stop_wait) {
	      grab_set_chan(naltch);
	      grab_free();
	      dsleep(0.1);
	      grab_lock();
	      grab_set_chan(nchan);
	   }
	}
	if(!no_tv_signal && (t-t0_add<GRAB_DELAY || nn<nadd)) {
	   grab_free();
	   dsleep((nchan!=naltch)?0.1:0.01);
	   continue;
	}
	t0_add = t0;

	stat_image(width, height);
	if(check_tv_signal() && auto_mode) {
	   adjust_video_control();
	   if(rbgd[nchan] && !color_mode) {   /* extract&remove bgd for TV-channel? */
	      if(nadd >= prev_nadd && do_bgd)
		 lin_col_bgd(width, height, width/2, height/2);
	      else
		 lin_col_bgd0();
	      prev_nadd = nadd;
	   }
	} else if(!no_tv_signal && nadd_mode && rbgd[nchan] && !color_mode)
	   lin_col_bgd(width, height, width/2, height/2);

	if(nchan!=naltch)
	   grab_set_chan(naltch);

	image = buff;
	out_image(image, width, height);
	make_mini_image(image, width, height, 3, buff_mini);
	add_text(image, width, height, NULL);
	add_text(buff_mini, width/3, height/3, str);
	if(no_tv_signal) {
	   add_text(image, width, height/3, "                  No TV signal!");
	   add_text(buff_mini, width/3, height/9, "     No TV!");
	   dsleep(1.5);
	} else {
	   sprintf( tmp,"Avr:%d",nadd);
	   add_text(image, width, CHAR_HEIGHT+3, tmp);
	}
	write_jpeg(jpeg_file_mini, buff_mini, width/3, height/3, JPEG_QUALITY);
	if(ftp_send)
	   ftp_upload(jpeg_file_mini, ftp_file_mini);

	write_jpeg(jpeg_file, image, width, height, JPEG_QUALITY);
	if(ftp_send)
	   ftp_upload(jpeg_file, ftp_file);

	grab_free();

	if(no_tv_signal) dsleep(GRAB_DELAY*2.0);
	else             dsleep((nchan!=naltch)?0.2:0.02);
    }
    if(ftp_send) FtpClose(ftp);
    return 0;
}

/*----VSher--------------------------------------------------------------------*/
static unsigned short img[GRAB_WIDTH*GRAB_HEIGHT];
static unsigned short imgc[GRAB_WIDTH*GRAB_HEIGHT*3];
static int hist[65536];
static int hbot=0,htop=65535;
static int vmin,vmax,med;              /* minimum, maximum, median     */
static double avr, dsp;                /* average level and dispersion */
#define BOTLEV (-4.)                   /* CUT-levels (in Sigma) */
#define TOPLEV (+7.)
static int vbot,vtop;                  /* real CUT-levels */

static void copy_image(unsigned char *image, int width, int height)
{
    int i,j,k;
    unsigned char *bp;
    unsigned short *sp, *spc;

    if(color_mode) {
       bp = image;
       sp = img;
       spc = imgc;
       for(j=0;j<height;j++)
	  for(i=0;i<width;i++) {
	     *sp = 0;
	     for(k=0;k<3;k++) {
		*spc++ = *bp;
		*sp += *bp++;
	     }
	     sp++;
	  }
    } else {
       bp = image;
       sp = img;
       for(j=0;j<height;j++)
	  for(i=0;i<width;i++)
	      *sp++ = *bp++;
    }
}

static void add_image(unsigned char *image, int width, int height)
{
    int i,j,k;
    unsigned char *bp;
    unsigned short *sp, *spc;

    if(color_mode) {
       bp = image;
       sp = img;
       spc = imgc;
       for(j=0;j<height;j++)
	  for(i=0;i<width;i++) {
	     for(k=0;k<3;k++) {
		*spc++ += *bp;
		*sp += *bp++;
	     }
	     sp++;
	  }
    } else {
       bp = image;
       sp = img;
       for(j=0;j<height;j++)
	  for(i=0;i<width;i++)
	      *sp++ += *bp++;
    }
}

static void out_image(unsigned char *image, int width, int height)
{
    int i,j,k,pix;
    unsigned char *bp;
    unsigned short *sp, *spc;

    if(color_mode) {
       bp = image;
       spc = imgc;
       for(j=0;j<height;j++)
	  for(i=0;i<width;i++) {
	     for(k=0;k<3;k++) {
		pix = (int)((double)(*spc-vbot)*255.0/(vtop-vbot)+0.5);
		if(pix>255) pix=255;
		if(pix<0) pix=0;
		*bp++ = pix;
		spc++;
	     }
	  }
    } else {
       bp = image;
       sp = img;
       for(j=0;j<height;j++)
	  for(i=0;i<width;i++) {
	     pix = (int)((double)(*sp-vbot)*255.0/(vtop-vbot)+0.5);
	     if(pix>255) pix=255;
	     if(pix<0) pix=0;
	     *bp++ = pix;
	     sp++;
	  }
    }
}

static void make_mini_image(unsigned char *image, int width, int height, int fact,unsigned char *mini_image)
{
    int i,ii,j,jj,k,pixi,pixo;
    int nc = color_mode? 3 : 1;
    int h = height/fact;
    int w = width/fact;

    for(j=0;j<h;j++) for(i=0;i<w;i++) for(k=0;k<nc;k++) {
       pixo=0;
       for(jj=0;jj<fact;jj++) for(ii=0;ii<fact;ii++) {
	  pixi = image[((j*fact+jj)*width + i*fact+ii)*nc + k];
	  if(pixi>pixo) pixo=pixi;
       }
       mini_image[(j*w+i)*nc+k] = pixo;
    }

  /*
   * for(j=0;j<h;j++)
   *    for(i=0;i<w;i++)
   *       for(k=0;k<nc;k++) {
   *          mini_image[(j*w+i)*nc+k] = image[(j*fact*width + i*fact)*nc+k];
   *    }
   */
}
static void prep_frame(unsigned char *image, int width, int height)
{/*single frame preparation for output, calculate: hbot,htop,vmin,vmax */
   int i,j,hd;
   int pix;
   unsigned char *bp;
   int nc = color_mode? 3 : 1;
   int wc = width*nc;
   int n = width*height;
   int nbot = (int)(n*0.1 + 0.5);     /*  ~10% of "bottom" pixels  */
   int ntop = (int)(n*0.9 + 0.5);     /*  ~10% of "top" pixels     */
   int notv = color_mode? 0xC0/3 : 0x23;

   for (j=0; j<256; j++) hist[j]=0;

   vmin = vmax = image[0];
   bp = image;
   for(j=0;j<height;j++)
      for(i=0;i<width;i++) {
	 pix = *bp++;
	 if(color_mode) { pix += *bp++; pix += *bp++; pix /= 3;}
	 hist[pix]++;             /* accumulation of histogram */
	 if(vmin>pix) vmin = pix;
	 if(vmax<pix) vmax = pix;
      }
   if(vmax==vmin && vmax==notv) { /* No TV-signal?!*/
      no_tv_signal=1;
      return;
   }
   no_tv_signal=0;
   hbot = 0;
   htop = 255;                   /* determine "limits" of histogram usage */
   for (i=j=0; j<256; j++) {     /* to reject ~20% top/bottom pixels      */
      i += hist[j];               /* for "robust" avr & dsp calculation    */
      if(i<=nbot) hbot = j;
      if(i<=ntop) htop = j;
   }
   while(htop-hbot < 120)
     { htop++; if(hbot>0) hbot--; }
/*printf("hbot=%d  htop=%d\n",hbot,htop);*/
   hd = htop-hbot;
   bp = image;
   for(j=0;j<height;j++)
      for(i=0;i<wc;i++) {
	 pix = (int)((double)(*bp-hbot)*200.0/(double)hd+0.5)+10;
	 if(pix>255) pix=255;
	 if(pix<0) pix=0;
	 *bp++ = pix;
      }
}
static void stat_image(int width, int height) /* do some statistics on img[] */
{                    /* calculate: hbot,htop,vmin,vmax,avr,med,dsp,vbot,vtop */
   int i,j,pix;
   int n,nbot,nmid,ntop;
   int mxcut,mncut;
   unsigned short *sp;

   n = width*height;
   nbot = (int)(n*0.05 + 0.5);     /*  ~5% of "bottom" pixels  */
   ntop = (int)(n*0.95 + 0.5);     /*  ~5% of "top" pixels     */
   nmid = n/2;

   for (j=0; j<65536; j++) hist[j]=0;
   vmin = vmax = img[0];
   sp = img;
   for(j=0;j<height;j++)
      for(i=0;i<width;i++) {
	 pix = *sp++;
	 hist[pix]++;             /* accumulation of histogram */
	 if(vmin>pix) vmin = pix;
	 if(vmax<pix) vmax = pix;
      }
   hbot = 0;
   med = 32767;
   htop = 65535;                  /* determine "limits" of histogram usage */
   for (i=j=0; j<65535; j++) {    /* to reject ~10% top/bottom pixels      */
      i += hist[j];               /* for "robust" avr & dsp calculation    */
      if(i<=nbot) hbot = j;
      if(i<=nmid) med  = j+1;     /* median */
      if(i<=ntop) htop = j+1;
    /*if(hist[j]!=0) printf("hist[%d]=%d i=%d\n",j,hist[j],i);*/
   }
 /*printf("mbot=%d  nmid=%d  ntop=%d  hbot=%d  med=%d  htop=%d\n",nbot,nmid,ntop,hbot,med,htop);*/
   for (avr=0.,i=0,j=hbot; j<=htop; j++) {
      avr += (double)hist[j]*j;
      i += hist[j];
   }
   if(i<1) i=1;
   avr /= i;
   for (dsp=0.,j=hbot; j<=htop; j++) {
      double d = j-avr;
      dsp += d*d*hist[j];
   }
   dsp = sqrt(dsp/i);
   vbot = (int)(avr+BOTLEV*dsp+0.5);
   vtop = (int)(avr+TOPLEV*dsp+0.5);
   mncut = hbot-(int)(dsp+0.5);
   mxcut = htop+(int)(dsp+0.5);
   if(mncut<vmin) mncut=vmin;
   if(mxcut>vmax) mxcut=vmax;
   if(vbot<mncut) vbot=mncut;
   if(vtop>mxcut) vtop=mxcut;
   if(color_mode) {
      hbot /= 3;  htop /= 3;
      vmin /= 3;  vmax = (vmax+2)/3; med = (med+1)/3;
      avr  /= 3.;  dsp /= 3.;
      vbot /= 3;  vtop /= 3;
   }
/*printf("Min=%d Max=%d Med=%d Avr=%.1f Sigma=%.2f\n",vmin,vmax,med,avr,dsp);*/

   while(vtop-vbot < 10*nadd)
     { vtop++; if(vbot>0) vbot--; }
/*printf("CUTs=%d,%d (Bot5%%=%d Top5%%=%d)\n",vbot,vtop, hbot,htop);*/
}
static int check_tv_signal() {
   int mn = vmin/nadd;
   int mx = vmax/nadd;
   int notv = color_mode? 0xC0/3 : 0x23;
   if(mx==mn && mx==notv) {
      fprintf(stderr,"No TV-signal?! ");
      grab_adjust_brightness(50.0-brightness);
      grab_adjust_contrast(30.0-contrast);
      if(nadd > 1) {
	 nadd = 1;
	 fprintf(stderr,"Set Nadd=%d  ",nadd);
      }
      no_tv_signal=1;
      fprintf(stderr,"\n");
      return(0);
   } else {
      no_tv_signal=0;
      return(1);
   }
}
static void adjust_video_control() {
   int mn0 = color_mode? 0 : 13;
   int mn1 = color_mode? 1 : 17;
   int mn2 = color_mode? 2 : 18;
   int mn3 = color_mode? 5 : 22;
   int mx0 = color_mode? 230 : 210;
   int mx1 = color_mode? 245 : 240;
   int mx2 = color_mode? 250 : 245;
   int mx3 = color_mode? 254 : 250;
   int mn = vmin/nadd;
 /*int mn = hbot/nadd;*/
   int mx = vmax/nadd;
 /*int mx = htop/nadd;*/
   int md = med/nadd;
   double av = avr/nadd;
   double ds = dsp/nadd;
   int dn = (nadd<11)? 1 : (nadd<21)? 2 : (nadd<41)? 5 : (nadd<81)? 10 : 20;
   int n;
/*printf("Min=%d Max=%d Med=%d Avr=%.1f Sigma=%.2f\n",mn,mx,md,av,ds);*/
   do_bgd = 0;
   if(mn<mn0)      grab_adjust_brightness( 5.);
   else if(mn<mn1) grab_adjust_brightness( 1.);
   else if(mn>mn3) grab_adjust_brightness(-5.);
   else if(mn>mn2) grab_adjust_brightness(-1.);

   if(av>128.) mx += ((int)av-128)/2;
   if(mx<mx0)      grab_adjust_contrast((md<mx0/4)? 7 : 3.);
   else if(mx<mx1) grab_adjust_contrast( 0.5);
   else if(mx>mx3) grab_adjust_contrast((md>mx0)? -7 : -3.);
   else if(mx>mx2) grab_adjust_contrast(-0.5);

   if( (ds<2.) &&
       ((color_mode && nadd==NADD_MAX_COLR) ||
	(!color_mode && nadd==NADD_MAX_GREY)))
      do_bgd = 1;

   n = nadd;
   if(ds>30.)
      nadd -= 50;
   else if(ds>20.)
      nadd -= 20;
   else if((md>180) || (md>150 && nadd>10) || (md>120 && nadd>20) || mx>mx3 ||
	(ds>5. && nadd>120.) || (ds>10. && nadd>50.) || (ds>15. && nadd>20.) )
      nadd -= dn;
   else if(ds<3.)
      nadd += 50;
   else if(ds<5.)
      nadd += 20;
   else if((md<30) || (md<60 && nadd<50) || (md<90 && nadd<20) || mx<120 ||
	   (ds<15. && nadd<20.) || (ds<10. && nadd<50.) )
      nadd += dn;
   if(color_mode) {
      if(nadd>NADD_MAX_COLR) nadd=NADD_MAX_COLR;
      if(nadd<NADD_MIN_COLR) nadd=NADD_MIN_COLR;
   } else {
      if(nadd>NADD_MAX_GREY) nadd=NADD_MAX_GREY;
      if(nadd<NADD_MIN_GREY) nadd=NADD_MIN_GREY;
   }
   if(nadd != n)
      fprintf(stderr,"Set Nadd=%d  ",nadd);
   fprintf(stderr,"\n");
}

static double lin[GRAB_WIDTH];
static double col[GRAB_HEIGHT];
/*
static double lbgd1[GRAB_WIDTH];
static double cbgd1[GRAB_HEIGHT];
static double lbgd2[GRAB_WIDTH];
static double cbgd2[GRAB_HEIGHT];
*/

static void lin_col_bgd0() {
   int i,j;
   for(i=0; i<GRAB_WIDTH; i++) lin[i] = 0.0;
   for(j=0; j<GRAB_HEIGHT; j++) col[j] = 0.0;

}

/* calc "line" & "column" bgd on img[] and subtruct it from */
static void lin1_col1_bgd(int width, int height)
{
   int i,j,k,n,mx,mn,pix;
   unsigned short *sp0,*sp;
   double av,s;

   sp0 = img;
   for(j=0; j<height; j++, sp0 += width) {
      av = avr;
      mx = vmax;
      mn = vmin;
      for(k=1;k<8;k++) {
	 sp = sp0;
	 s=0.;
	 for(n=i=0; i<width; i++) {
	    pix = *sp++;
	    if(mn<=pix && pix<=mx) {
	       s += pix; n++;
	    }
	 }
	 if(n==0) n=1;
	 av = s/n;
	 mx = (int)(av+1.*dsp/k+1.5);
	 mn= (int)(av-7.*dsp/k-1.5);
      }
    /*col[j] = col[j]*0.5 + av*0.5;*/
      col[j] = av;
   }
   sp0 = img;
   for(i=0; i<width; i++, sp0++) {
      av = avr;
      mx = vmax;
      mn = vmin;
      for(k=1;k<8;k++) {
	 sp = sp0;
	 s=0.;
	 for(n=j=0; j<height; j++, sp += width) {
	    pix = *sp;
	    if(mn<=pix && pix<=mx) {
	       s += pix; n++;
	    }
	 }
	 if(n==0) n=1;
	 av = s/n;
	 mx = (int)(av+1.*dsp/k+1.5);
	 mn= (int)(av-7.*dsp/k-1.5);
      }
    /*lin[i] = lin[i]*0.4 + av*0.6;*/
      lin[i] = av;
   }
 /*
  * dline1_bgd(lin, width,  100, lbgd1);
  * dline1_bgd(col, height, 100, cbgd1);
  * dline1_bgd(lin, width,  10, lbgd2);
  * dline1_bgd(col, height, 10, cbgd2);
  */
   s = 0.0;
   for(i=0; i<width; i++) s += lin[i];
   av = s/width;
   s = 0.0;
   for(j=0; j<height; j++) s += col[j];
   av += s/height;
   sp = img;
   for(j=0; j<height; j++)
      for(i=0; i<width; i++,sp++) {
      /* double bg = (lbgd2[i]+cbgd2[j]) - (lbgd1[i]+cbgd1[j]); */
	 *sp -= (int)(lin[i]+col[j]-av+0.5);
      }
}

static double dbuf[GRAB_WIDTH+GRAB_HEIGHT];

static void dline1_bgd(double *line, int len, int wnd, double *bgd) {
   int ni=7;
   int i,j,l,k,n;
   double s;
   for(i=0; i<len; i++) dbuf[i]=line[i];
   for(k=1;k<ni;k++) {
      s=0.0;
      for(i=0; i<wnd/2; i++) s += dbuf[i];
      for(j=0,n=i; i<wnd; i++,j++) {
	 s += dbuf[i]; n++;
	 bgd[j] = s/n;
      }
      for(l=0; j<len; i++,j++,l++) {
	 if(i<len)
	    s += dbuf[i]-dbuf[l];
	 else {
	    s -= dbuf[l]; n--;
	 }
	 bgd[j] = s/n;
      }
      if(k<ni-1)
	 for(i=0; i<len; i++)
	    if(fabs(dbuf[i]-bgd[i]) > 1.0*dsp)
	       dbuf[i] = bgd[i];
   }
}

static int ibuf[GRAB_WIDTH+GRAB_HEIGHT];

static void iline1_bgd(unsigned short *line, int len, int wnd, int *ibgd) {
   int ni=7, idsp=dsp+0.5;
   int i,j,l,k,n;
   int s;
   for(i=0; i<len; i++) ibuf[i]=line[i];
   for(k=1;k<ni;k++) {
      s=0;
      for(i=0; i<wnd/2; i++) s += ibuf[i];
      for(j=0,n=i; i<wnd; i++,j++) {
	 s += ibuf[i]; n++;
	 ibgd[j] = (s+n/2)/n;
      }
      for(l=0; j<len; i++,j++,l++) {
	 if(i<len)
	    s += ibuf[i]-ibuf[l];
	 else {
	    s -= ibuf[l]; n--;
	 }
	 ibgd[j] = (s+n/2)/n;
      }
      idsp=dsp/k+0.5;
      if(k<ni-1) {
	 for(i=0; i<len; i++) {
	    if(ibuf[i]-ibgd[i] > idsp)
	       ibuf[i] = ibgd[i];
	    else if(ibuf[i]-ibgd[i] < -10*idsp)
	       ibuf[i] = ibgd[i];
	 }
      }
   }
}
static unsigned short img_sav[GRAB_WIDTH*GRAB_HEIGHT];
static int ibgd[GRAB_WIDTH+GRAB_HEIGHT];
static unsigned short ibf[GRAB_WIDTH+GRAB_HEIGHT];

/* extract every line & column bgd from img[] (and subtruct it) */
static void lin_col_bgd(int width, int height, int hwnd, int vwnd)
{
   int i,j;
   unsigned short *sp0,*sp;
   int iav=avr+0.5;
   int hbot_s,htop_s,vmin_s,vmax_s,med_s,vbot_s,vtop_s;
   double avr_s, dsp_s;

   sp0 = img;
   sp = img_sav;
   for(j=0; j<height; j++)
      for(i=0; i<width; i++)
	 *sp++ = *sp0++;
   sp0 = img;
   for(i=0; i<width; i++,sp0++) {
      for(sp=sp0,j=0; j<height; j++,sp+=width) ibf[j] = *sp;
      iline1_bgd(ibf,height,vwnd,ibgd);
      for(sp=sp0,j=0; j<height; j++,sp+=width) *sp -= ibgd[j] - iav;
   }
   sp = img;
   for(j=0; j<height; j++) {
      iline1_bgd(sp,width,hwnd,ibgd);
      for(i=0; i<width; i++,sp++) *sp -= ibgd[i] - iav;
   }
   hbot_s = hbot; htop_s = htop;
   vmin_s = vmin; vmax_s = vmax;
    med_s =  med;
    avr_s =  avr;  dsp_s =  dsp;
   vbot_s = vbot; vtop_s = vtop;

   stat_image(width, height);
/*printf("BGD: Avr:%.1f->%.1f  Sigma:%.2f->%.2f\n",avr_s/nadd,avr/nadd,dsp_s/nadd,dsp/nadd);
 */
   if(dsp > 0.25*dsp_s || dsp/nadd>0.5) {
      sp0 = img;
      sp = img_sav;
      for(j=0; j<height; j++)
	 for(i=0; i<width; i++)
	    *sp0++ = *sp++;
      hbot = hbot_s; htop = htop_s;
      vmin = vmin_s; vmax = vmax_s;
       med =  med_s;
       avr =  avr_s;  dsp =  dsp_s;
      vbot = vbot_s; vtop = vtop_s;
   }
}

static void dsleep(double dt)
{
  int nfd;
  struct timeval tv;
   tv.tv_sec = (int)dt;
   tv.tv_usec = (int)((dt - tv.tv_sec)*1000000.);
slipping:
   nfd = select(0, (fd_set *)NULL,(fd_set *)NULL,(fd_set *)NULL, &tv);
   if(nfd < 0) {
      if(errno == EINTR)
	/*On  Linux,  timeout  is  modified to reflect the amount of
	 time not slept; most other implementations DO NOT do this!*/
	 goto slipping;
      fprintf(stderr,"Error in mydelay(){ select() }. %s\n",strerror(errno));
   }
}

static double dtime() {
   struct timeval ct;
   struct timezone tz;
   gettimeofday(&ct, &tz);
   return (ct.tv_sec + ct.tv_usec/1e6);
}

